// Terrain Editor
// Marielle Lange - mlange@widged.com
/*
  Adapted from LandscapeMapEditor on wondefl.
  http://wonderfl.net/code/3f0c5718a78a0d4adf9e7b2d2ddbd67cc328ac70

  MIT License - http://www.opensource.org/licenses/mit-license.php
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.
*/

package {

   import com.bit101.components.*;
   
   import flash.display.*;
   import flash.events.*;
   import flash.filters.*;
   import flash.geom.*;

   [SWF(width="600", height="450", backgroundColor="#FFFFFF", frameRate="20")]
    public class TerrainEditor extends Sprite {

      private var _key:KeyMapper;
      
      private var w:int = 400;
      private var h:int = 400;

      private var render:Bitmap; 
      private var texture:BitmapData;
      private var heightMap:BitmapData; 
      private var editingZone:Shape; 

      private var mapRGB:Array = new Array(3);
      
      private var size:int = 100;
      private var add:Boolean = true;
      private var random:Boolean = true;

      // terrain parameters
      private var perlinRandomSeed:int = 1000  // Math.random()*0xffffffff; // fix number for reproducible results
      private var perlinFreq:Number = 0.5;
      private var perlinOctaveQty:int = 10;
      // -- height map
      private var greyMultiplier:Number = 1.5; // increases the contrast, that is creates clear boundaries between water (darker gray) and land (lighter gray). 
      private var greyOffset:int = -64;        // defines the range of gray values.  Too low and there is no mauntain. Too high and there is no water
      private var isColorized:Boolean = true;
      // shading
      private var shadingMultiplier:Number = 4;
      private var shadingOffset:int = 160;
      private var shadingConvolutionFilter:ConvolutionFilter = new ConvolutionFilter(3,3,[-3,-2,0,-2,0,2,0,2,3],3,1,true,true);
          // before fork: cnew ConvolutionFilter(3,3,[-1,-1,0,-1,0,1,0,1,1],1,0,true,true);

      public function TerrainEditor() {
         initializeEvents();
         initializeMap();
         prepareMap();         
         renderMap();

         initializeControlPanel();
      }
    
       private function initializeEvents():void
       {
            var o:Object= new Object();
            o[KeyMapper.UP]               = "increaseRadius";
            o[KeyMapper.DOWN]             = "decreaseRadius";
            o[KeyMapper.ARROW_UP]         = "hideRadius";
            o[KeyMapper.MOUSE_DOWN]       = "addTerrain";
            o[KeyMapper.MOUSE_DOWN_SHIFT] = "removeTerrain";
            o[KeyMapper.SPACE]            = "toggleRandom";
            addEventListener("increaseRadius", onIncreaseRadius);
            addEventListener("decreaseRadius", onDecreaseRadius);
            addEventListener("hideRadius", onHideRadius);
            addEventListener("addTerrain", onAddTerrain);
            addEventListener("removeTerrain", onRemoveTerrain);
            addEventListener("toggleRandom", onToggleRandom);
            _key = new KeyMapper(this, o);
       }
       
       // ##################
       //  Events, Interactivity
       // ##################
       private function onIncreaseRadius(event:Event):void
       {
           size += 10;
           refreshEditingZone();
         }
       private function onDecreaseRadius(event:Event):void
       {
           size -= 10;
           refreshEditingZone();
       }
 
       private function onAddTerrain(event:Event):void
       {
          add = true;
          updateMap();
       }
       private function onRemoveTerrain(event:Event):void
       {
          add = false;
          updateMap();
       }
       
       private function onToggleRandom(event:Event):void
       {
          random = !random;
       }
       private function onHideRadius(event:Event):void
       {
         clearEditingZone();
       }
      
       // ##################
       //  Terrain
       // ##################
      private function initializeMap():void
      {
         // rendering of the map, what will be modified thereafter is the texture data
         texture = new BitmapData(w, h, false, 0);
         render = new Bitmap(texture);
         addChild(render);
   
         // visual feedback on where the editing zone is
         editingZone = new Shape();
         addChild(editingZone);
      }
      
      private function prepareMap():void
      {
         mapRGB = getMapColors();              // gradientBar
         heightMap = generateHeightMap(w,h);   // height map
      } 
      
      /**
      * Draws a bar of 256x1 with terrain colors and transitions 
      * (blue, light brown green, dark brown, white)
      */
      private function getMapColors():Array
      {
         var mat:Matrix = new Matrix();
         mat.createGradientBox(256,1,0,0,0);

         var gradation:Object = {
             color:[0x000080,0x0066ff,0xcc9933,0x00cc00,0x996600,0xffffff], 
             alpha:[100, 100, 100, 100, 100, 100], 
             ratio:[0, 96, 96, 128, 168, 224]
         };
         var gradientBar:Shape = new Shape();
         var g:Graphics = gradientBar.graphics;
         g.clear();
         g.beginGradientFill("linear", gradation.color, gradation.alpha, gradation.ratio, mat);
         g.drawRect(0,0,256,1);
         g.endFill();
 
         mat.identity();

         var mapR:Array=new Array(256);
         var mapG:Array=new Array(256);
         var mapB:Array=new Array(256);
         // initialize the RGB arrays
         var gmap:BitmapData = new BitmapData(256,1,false,0); 
         gmap.draw(gradientBar);
         for (var i:int=0; i<256; i++) {
             var col:uint = gmap.getPixel(i, 0);
             // get the red,green,blue channel values for each pixel in the gradient bar 
             // i then stand for the level of grey intensity in the perlin noise image
             mapR[i] = col & 0xff0000;  // R  
             mapG[i] = col & 0x00ff00;  // G
             mapB[i] = col & 0x0000ff;  // B
         }
         gmap.dispose();

         return [mapR, mapG, mapB];           
      }

      /**
      * Draws perlin noise that is meant to represent variation in height
      */
      private function generateHeightMap(w:int, h:int):BitmapData
      {
         var bdata:BitmapData = new BitmapData(w, h, false, 0); // heightmap
         // Generates a Perlin noise image.
         // perlinNoise(baseX, baseY, numOctaves, randomSeed, stitch, fractalNoise, channelOptions, grayScale, offsets)
         // Larger numbers of octaves create images with greater detail. Larger numbers of octaves also require more processing time.
         // stitch If the value is true, the method attempts to smooth the transition edges of the image to create seamless textures for tiling as a bitmap fill.
         // fractalNoise - If true, the method generates fractal noise; otherwise, it generates turbulence. An image with turbulence has visible discontinuities in the gradient that can make it better approximate sharper visual effects like flames and ocean waves.
         bdata.perlinNoise(w*perlinFreq, h*perlinFreq, perlinOctaveQty, perlinRandomSeed, false, false, 0, true);
         // Adjusts the color values in a specified area of a bitmap image by using a 
         // ColorTransform object. 
         // colorTransform(redMultiplier, greenMultiplier, blueMultiplier, alphaMultiplier, redOffset, greenOffset, blueOffset, alphaOffset)
         // [color]Offset: A number from -255 to 255 that is added to the [color] channel value after it has been multiplied by the [color]Multiplier value.
         //  multiplier and offset need to be identical for all channels to keep a greyscale image.
         bdata.colorTransform(bdata.rect, new ColorTransform(greyMultiplier, greyMultiplier, greyMultiplier, 1, greyOffset, greyOffset, greyOffset, 0));
         return bdata;
      }
      
       
      /**
      * Edit the map to add or remove terrain
      */
      private function updateMap():void
      {
         // zone that will be updated
         refreshEditingZone();
         // update heigthmap
          if(add){
              heightMap.draw(editingZone,null,null,BlendMode.ADD);
          } else {
              heightMap.draw(editingZone,null,null,BlendMode.SUBTRACT);
          }
         renderMap();
         // clear out the zone
         clearEditingZone();
      }
      
      private function renderMap():void
      {
         // Turns the perlin noise from black and white to a version colorized with the gradient bar.
         // paletteMap(bmpData, sourceRect, destPoint, redArray, greenArray, blueArray, alphaArray)
         // Remaps the color channel values in an image that has up to four arrays of color palette data, one for each channel.
         // texture.paletteMap(heightMap, heightMap.rect, heightMap.rect.topLeft);  
         if(isColorized)
            texture.paletteMap(heightMap, heightMap.rect, heightMap.rect.topLeft, mapRGB[0], mapRGB[1], mapRGB[2]);
         else
            texture.paletteMap(heightMap, heightMap.rect, heightMap.rect.topLeft);

         // shading map
         var shadingMap:BitmapData= new BitmapData(w, h, false, 0); 
         shadingMap.applyFilter(heightMap, heightMap.rect, heightMap.rect.topLeft, shadingConvolutionFilter);
         var colorTransform:ColorTransform = new ColorTransform(shadingMultiplier, shadingMultiplier, shadingMultiplier, 1, shadingOffset, shadingOffset, shadingOffset, 0);    
	
	      // Draws the source display object onto the bitmap image, using the Flash Player or AIR vector renderer.
         // draw(bmpData, matrix, colorTransform, blendMode, clipRect, smoothing)
         texture.draw(shadingMap, null, colorTransform, BlendMode.MULTIPLY);
      }


      private function refreshEditingZone():void
      {
         var g:Graphics =  editingZone.graphics;
         clearEditingZone();
         addCircle(g, new Point(render.mouseX, render.mouseY), size, add);
      }

     /**
      * Add a circle that marks the zone where terrain will be added or removed. 
      * @param int coord the location on screen where the circle shold appear
      * @param boolean add  if true, terrain is added. Otherwise terrain is removed.  
      * @param int r the circle radius
      */
      private function addCircle(g:Graphics, coord:Point, r:int, add:Boolean):void 
      {
         var x:int = coord.x;
         var y:int = coord.y;
          if(random) {
              x += Math.random()*r*2-r;
              y += Math.random()*r*2-r;
          }
          g.beginFill(0x010101);
          g.drawCircle(x, y, r);
          g.endFill();
      }
      
     /**
      * Add a circle that marks the zone where terrain will be added or removed. 
      */
      private function clearEditingZone():void
      {
          editingZone.graphics.clear();
         
      }

       // ##################
       //  Control Panel
       // ##################
        private function initializeControlPanel():void
        {
           var rowY:int;
           var rowHeight:int;
           var bottomPanel:Panel = new Panel(this, 0, 400);
           bottomPanel.setSize(400, 50);
           rowHeight = 12;
           rowY = 3;
           new Label(bottomPanel, 7, rowY, "mouse down to add terrain - shift+mouse down to remove terrain");
           rowY += rowHeight; 
           new Label(bottomPanel, 7, rowY, "left arrow to decrease editing circle - right arrow to increase it"); 
           rowY += rowHeight; 
           new Label(bottomPanel, 7, rowY, "space to toggle the random change in circle position"); 
           
           rowHeight = 20; 
           rowY = 7;
           var rightPanel:Panel = new Panel(this, 400, 0);
           rightPanel.setSize(200, 400);


         // grey height map
          new Label(rightPanel, 7, rowY, "greyMultiplier:"); 
          var greyM:InputText = new InputText(rightPanel, 100, rowY, greyMultiplier.toString(),
                function( event:Event ):void {
                   greyMultiplier = parseFloat(greyM.text);
                   prepareMap(); renderMap();
               } 
              );
           greyM.setSize(80, 16);
         
          rowY += rowHeight; 
          new Label(rightPanel, 7, rowY, "greyOffset:"); 
          var greyO:InputText = new InputText(rightPanel, 100, rowY, greyOffset.toString(),
                function( event:Event ):void {
                   greyOffset = parseInt(greyO.text);
                   prepareMap(); renderMap();
               } 
              );
           greyO.setSize(80, 16);
         
           // colorization
          rowY += rowHeight + 5; 
           var colorCheck:CheckBox = new CheckBox(rightPanel, 7, rowY, "colorize", 
            function(e:Event):void { isColorized = colorCheck.selected; renderMap(); }
            )
            colorCheck.selected = isColorized;



          // shading
           rowY += rowHeight + 5; 
           new Label(rightPanel, 7, rowY, "shadingMultiplier:"); 
           var shadingM:InputText = new InputText(rightPanel, 100, rowY, shadingMultiplier.toString(),
                 function( event:Event ):void {
                    shadingMultiplier = parseFloat(shadingM.text);
                    renderMap();
                } 
               );
            shadingM.setSize(80, 16);

           rowY += rowHeight; 
           new Label(rightPanel, 7, rowY, "shadingOffset:"); 
           var shadingO:InputText = new InputText(rightPanel, 100, rowY, shadingOffset.toString(),
                 function( event:Event ):void {
                    shadingOffset = parseInt(shadingO.text);
                    renderMap();
                } 
               );
            shadingO.setSize(80, 16);


        }
    }
}

